# STICK TO SEMANTIC VERSIONING:
# https://semver.org/
# TLDR:
# Given a version number MAJOR.MINOR.PATCH, increment the:
#
# MAJOR version when you make incompatible API changes,
# MINOR version when you add functionality in a backwards compatible manner, and
# PATCH version when you make backwards compatible bug fixes.
__version__ = "3.6.2"
